
/**CFile**************************************************************************
  FileName    [bcdEL.c]
  PackageName [bcd]
  Synopsis    [Routines to detect bad cycles based on Emerson-Lei's method
  Author      [Aiguo Xie]
  Date        [10/3/1999]
  Copyright   [Copyright (c) 1999 Aiguo Xie. All rights reserved]
*********************************************************************************/

#include "fsmInt.h"
#include "bdd.h"
#include "ntm.h"
#include "img.h"
#include "imgInt.h"
#include "sccInt.h"
#include "bcdInt.h"

static void bcd_EL_Init(Fsm_Fsm_t *fsm);

int bcd_EL(Fsm_Fsm_t *fsm, mdd_t *reachable, sccMddList *csList)
{
    Img_ImageInfo_t *imageInfo;
    sccMddList *csCmplList, *temp_cur, *listCur;
    mdd_t *b, *new_b, *fi, *fi_star, *star_fi, *f_pound;
    mdd_t *plus_b, *b_plus;
    int change, iter;

    bcd_EL_Init(fsm);

    iter = 0;

    imageInfo = fsm->imageInfo;

    /* complementing the cycle sets */
    csCmplList = bcd_cmplList(fsm, reachable, csList);

    /* the main algorithm */

    /* step 1: */
    b = mdd_dup(reachable);

    fprintf(vis_stderr, "ITERATION: ");
    do {
	fprintf(vis_stderr, "%d ", ++iter);
        change = 0;
        /* step 2: */
        listCur = csCmplList;
        while (listCur) {
            listCur->val = scc_MddAndDestroy(listCur->val, b, 1, 1, 1, 0);
            fi = listCur->val;
            fi_star = scc_ForwardReachableStates(fsm, fi, reachable);
            star_fi = scc_BackwardReachableStates(fsm, fi,reachable);
	    if (listCur == csCmplList)
		f_pound = fi_star;
	    else
            	f_pound = scc_MddAndDestroy(f_pound, fi_star,1,1,1,1);
            f_pound = scc_MddAndDestroy(f_pound, star_fi,1,1,1,1);
            listCur = listCur->next;
        }
        /* temp 3: */
        new_b = f_pound;
	while (bcd_mddsAreNotEqual(b, new_b)) {
            change++;
            mdd_free(b);
	    b = new_b;
            b_plus = Img_ImageInfoComputeFwdWithDomainVars(imageInfo,b,b,b);
            plus_b = Img_ImageInfoComputeBwdWithDomainVars(imageInfo,b,b,b);
	    
            new_b = scc_MddAndDestroy(b, b_plus, 1, 1, 0, 1);
            new_b = scc_MddAndDestroy(new_b, plus_b, 1, 1, 1, 1);
        } 
    } while (change); /* step 4: */
    fprintf(vis_stderr, "\n");
    /* step 5: */
    if (mdd_is_tautology(new_b, 0))
	fprintf(vis_stderr, "vis detects no bad cycle.\n");
    else
	fprintf(vis_stderr, "vis has detected some bad cycle.\n");

    /* clean up before leave */
    mdd_free(b);
    mdd_free(new_b);
    listCur = csCmplList;
    while (listCur) {
        temp_cur = listCur->next;
        mdd_free(listCur->val);
        free(temp_cur);
        listCur = temp_cur;
    }
    return 0;
}

sccMddList *bcd_cmplList(Fsm_Fsm_t *fsm, mdd_t *space, sccMddList *list) {
sccMddList *lastListElement, *newListElement, *mddListCurs, *cmplList;
mdd_t *tempMdd;

    cmplList = NULL;
    mddListCurs = list;
    lastListElement = NULL;
    while (mddListCurs) {
        newListElement = (sccMddList *)calloc(1,sizeof(sccMddList));

        tempMdd = mdd_not(mddListCurs->val);
        newListElement->val = mdd_and(tempMdd, space,1,1);
        mdd_free(tempMdd);

	newListElement->id = mddListCurs->id;
    	newListElement->size = bcd_numStates(fsm, newListElement->val);
        if (cmplList == NULL) {
            cmplList = newListElement;
        }
        if (lastListElement)
            lastListElement->next = newListElement;
        lastListElement = newListElement;
        mddListCurs = mddListCurs->next;
    }
    return cmplList;
}

int bcd_mddsAreNotEqual(mdd_t *d1, mdd_t *d2)
{
    mdd_t *temp;

    temp = mdd_and(d1, d2, 1, 0);
    if (mdd_is_tautology(temp, 0)) {
        mdd_free(temp);
        return 0;
    } else {
        mdd_free(temp);
        return 1;
    }
    /* never reached */
}

static void bcd_EL_Init(Fsm_Fsm_t *fsm)
{
    mdd_manager *mddManager;

    fsm->imageInfo = Fsm_FsmReadOrCreateImageInfo(fsm,1,1);

    if (bcdBddCacheFlush == BCD_TRUE) {
    	mddManager = 
    	Part_PartitionReadMddManager(fsm->imageInfo->functionData.mddNetwork);
	fprintf(vis_stderr, "flushing bdd cache...");
    	cuddCacheFlush(mddManager);
	fprintf(vis_stderr, "done.\n");
    }
}
