/**CFile***********************************************************************

  FileName    [util.c]

  PackageName [bcd]

  Synopsis    [Utilities for bad cycle detection and scc enumeration]

  Author      [Aiguo Xie]

  Copyright   [Copyright (c) 1999, Aiguo Xie]

******************************************************************************/

#include "mcInt.h" 
#include "sccInt.h"
#include "bcdInt.h"

/*---------------------------------------------------------------------------*/
/* Macro declarations                                                        */
/*---------------------------------------------------------------------------*/

/**Function********************************************************************

  Synopsis [Pick a minterm from the given set with a maximal Hamming_Distance
	    from a given minterm]

  SideEffects []

******************************************************************************/
mdd_t *
bcd_Compute_A_Maximal_Hamming_Distance_Minterm( 
  mdd_t *aSet,
  mdd_t *aMinterm,
  array_t *Support,
  mdd_manager *mddMgr)
{
  int i, j;
  mdd_t *aSetDup;
  mdd_t *resultMdd;
  array_t *resultMddArray = array_alloc( mdd_t *, 0 );
  

  /*
   * check support of set is contained in Support (only in debugged executables)
   */
  assert( SetCheckSupport( aSet, Support, mddMgr )); 
  assert( SetCheckSupport( aMinterm, Support, mddMgr )); 

  aSetDup = mdd_dup( aSet );
  for( i = 0 ; i < array_n( Support ); i++ ) {

    int aSupportVar = array_fetch( int, Support, i );

    for( j = 0 ;;) {

      mdd_t *faceMdd, *tmpMdd_1, *tmpMdd_2, *tmpMdd_1_bar;

      array_t *tmpArray = array_alloc( int, 0 );
      array_insert_last( int, tmpArray, j );

      tmpMdd_1_bar = NULL;

      /* this call will crash if have run through range of mvar */
      faceMdd = mdd_literal( mddMgr, aSupportVar, tmpArray );
      array_free( tmpArray );

      tmpMdd_1 = mdd_and(faceMdd, aMinterm, 1, 1);
      mdd_free(faceMdd);
      if (!mdd_is_tautology(tmpMdd_1, 0)) {
	  tmpMdd_1_bar = mdd_not(tmpMdd_1);
	  mdd_free(tmpMdd_1);
      }

      if (tmpMdd_1_bar)
      	tmpMdd_2 = mdd_and( tmpMdd_1_bar, aSetDup, 1, 1 );
      else
	tmpMdd_2 = mdd_and( tmpMdd_1, aSetDup, 1, 1);

      if ( !mdd_is_tautology( tmpMdd_2, 0 ) ) {
	if (tmpMdd_1_bar)
        	array_insert_last( mdd_t *, resultMddArray, tmpMdd_1_bar );
	else
		array_insert_last( mdd_t *, resultMddArray, tmpMdd_1);
        mdd_free( aSetDup );
        aSetDup = tmpMdd_2;
        break;
      } else {
	mdd_free( tmpMdd_1);
	if (tmpMdd_1_bar)
      		mdd_free( tmpMdd_1_bar ); 
      	mdd_free( tmpMdd_2 );
      }
      j++;
    }
  }
  mdd_free( aSetDup );

  resultMdd = mdd_one( mddMgr );
  for ( i = 0 ; i < array_n( resultMddArray ); i++ ) {
    mdd_t *faceMdd = array_fetch( mdd_t *, resultMddArray, i );
    mdd_t *tmpMdd = mdd_and( faceMdd, resultMdd, 1, 1 );
    mdd_free( resultMdd ); mdd_free( faceMdd );
    resultMdd = tmpMdd;
  }
  array_free( resultMddArray );

  return resultMdd;
}

/**Function********************************************************************

  Synopsis [Select a state from the given set with a maximal Hamming distance
	    from a given state]

  SideEffects []

******************************************************************************/
mdd_t *
bcd_Compute_A_Maximal_Hamming_Distance_State( 
  mdd_t *states,
  mdd_t *from_state,
  Fsm_Fsm_t *modelFsm)
{
  array_t *PSVars = Fsm_FsmReadPresentStateVars( modelFsm );
  mdd_manager *mddMgr = Ntk_NetworkReadMddManager( Fsm_FsmReadNetwork( modelFsm ) ); 
  mdd_t *result = bcd_Compute_A_Maximal_Hamming_Distance_Minterm( states, from_state, PSVars, mddMgr );

  return result;
}
